import logging
import requests
import time
import docker
import asyncio
import sys

import rov_server_config
import utils

from concurrent.futures import ThreadPoolExecutor

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S')

health_instances = []


def server_health_check(client, instance_id):
    """
    服务健康检查
    :param client: 客户端连接
    :param instance_id: instanceId
    :return:
    """
    global health_instances
    ip_address = utils.get_ip_address(client=client, instance_id=instance_id)
    i = 0
    while i < 7:
        health_check_url = f"http://{ip_address}:5001/healthcheck"
        try:
            http_code = requests.get(health_check_url).status_code
        except:
            logging.info("images is downloading")
            http_code = 0

        if http_code == 200:
            health_instances.append((instance_id, ip_address))
            break
        else:
            time.sleep(10)
            i += 1


def restart(slb_client, instance_id, image_name):
    """重启容器并探活"""
    # 1. 获取ip
    ip_address = utils.get_ip_address(client=slb_client, instance_id=instance_id)
    logging.info(f"ip address：{instance_id}/{ip_address}")
    # 3. 重启容器
    docker_client = docker.DockerClient(base_url=f'tcp://{ip_address}:2375', timeout=60)
    image_id = docker_client.containers.get(image_name)
    image_id.restart()
    logging.info(f"docker restart finished, ip address：{instance_id}/{ip_address}")
    time.sleep(5)
    # 4. 探活
    server_health_check(slb_client, instance_id)
    logging.info(f"health check finished, ip address：{instance_id}/{ip_address}")


async def server_restart_all(slb_client, instance_id_list, image_name, max_workers):
    try:
        # 1. 摘流量
        utils.set_instance_weight_process(client=slb_client,
                                          slb_id=rov_server_config.slb_id,
                                          instance_id_list=instance_id_list,
                                          weight_list=[(0, 60)])
        logging.info(f"set weight with 0 finished.")

        # 2. 异步重启 & 探活
        global health_instances
        loop = asyncio.get_running_loop()
        executor = ThreadPoolExecutor(max_workers=max_workers)
        tasks = [
            loop.run_in_executor(executor, restart, *args) for args in
            [(slb_client, instance_id, image_name) for instance_id in instance_id_list]
        ]
        await asyncio.wait(tasks)
        logging.info(f"health instances count: {len(health_instances)}, {health_instances}")

        # 3. 挂载流量
        if len(health_instances) == len(instance_id_list):
            time.sleep(30)
            add_weight_list = [(10, 30), (20, 20), (40, 10), (60, 10), (80, 10), (100, 10)]
            utils.set_instance_weight_process(client=slb_client,
                                              slb_id=rov_server_config.slb_id,
                                              instance_id_list=[instance[0] for instance in health_instances],
                                              weight_list=add_weight_list)
            logging.info(f"set weight finished instance count: {len(health_instances)}, {health_instances}")
        elif len(health_instances) / len(instance_id_list) > 0.9:
            time.sleep(30)
            add_weight_list = [(10, 30), (20, 20), (40, 10), (60, 10), (80, 10), (100, 10)]
            utils.set_instance_weight_process(client=slb_client,
                                              slb_id=rov_server_config.slb_id,
                                              instance_id_list=[instance[0] for instance in health_instances],
                                              weight_list=add_weight_list)
            logging.info(f"set weight finished instance count: {len(health_instances)}, {health_instances}")

        else:
            logging.info(f"ess count: {len(instance_id_list)}, "
                         f"health count: {len(health_instances)}")
            sys.exit()

    except Exception as e:
        logging.error(f"server restart fail!")
        logging.error(e)


def main():
    try:
        slb_client = utils.connect_client(access_key_id=rov_server_config.slb_client_params['access_key_id'],
                                          access_key_secret=rov_server_config.slb_client_params['access_key_secret'],
                                          region_id=rov_server_config.slb_client_params['region_id'])

        # 1. 获取slb下所有机器
        online_instance_ids = utils.get_instance_ids(client=slb_client, slb_id=rov_server_config.slb_id)
        online_instance_count = len(online_instance_ids)
        logging.info(f"online instance count: {online_instance_count}.")
        logging.info(f"online instance ids: {online_instance_ids}")

        # 2. 重启
        image_name = 'rov-server'
        asyncio.run(server_restart_all(slb_client=slb_client, instance_id_list=online_instance_ids,
                                       image_name=image_name, max_workers=online_instance_count))
        logging.info(f"server restart finished!")
    except Exception as e:
        logging.error(e)


if __name__ == '__main__':
    main()
